"""
Everything related to support older Provider API versions
by translating them to the latest supported version.
"""


def translate_v0_2_to_v0_4(data):
    # The only two noticeable changes from our point of view are:
    # - timestamps converted from floating-point seconds to milliseconds;
    # - "trip_ids" now is a single "trip_id"
    for status_change in data["status_changes"]:
        # We were already expecting milliseconds in the 0.2 implementation
        if "." in str(status_change["event_time"]):
            status_change["event_time"] = round(status_change["event_time"] * 1000)
        # Keep only the first (and probably only) trip ID
        associated_trip = status_change.pop("associated_trips", None)
        if associated_trip:
            status_change["associated_trip"] = associated_trip[0]

    # Now up to date
    return data



"""
Pulling data for registered providers

This is the opposite of provider pushing their data to the agency API.
"""
import datetime
import enum
import logging
import urllib.parse
import uuid

from django.conf import settings
from django.contrib.gis import geos
from django.db import transaction
from django.utils import timezone
from django.utils.dateparse import parse_duration

import requests
from retrying import retry

from mds import db_helpers
from mds import enums
from mds import models
from mds import utils
from mds.provider_mapping import PROVIDER_REASON_TO_AGENCY_EVENT
from .oauth2_store import OAuth2Store
from .translation import translate_v0_2_to_v0_4


MDS_CONTENT_TYPE = "application/vnd.mds.provider+json"


logger = logging.getLogger(__name__)


# polling_cursor configuration field translated to query parameter
class POLLING_CURSORS(enum.Enum):
    start_time = "start_time"  # from the MDS specification
    start_recorded = "start_recorded"  # vendor only, 0.3 only
    total_events = "skip"  # vendor only, 0.3 only


class StatusChangesPoller:
    """Poll the given provider to fetch its latest data.

    Polling is automatically resumed from what was last saved in the provider config,
    unless cursor parameters are given.

    The from and to cursor values depend on the cursor type:
    - integer timestamp in milliseconds for "start_recorded" and "start_time"
    - integer count of lines for "total_events" (think LIMIT in SQL)

    Args:
        provider: Provider, the provider to poll
        cursor: POLLING_CURSORS, cursor type
        from_cursor: timestamp or int, lower limit
        to_cursor: timestamp or int, upper limit
    """

    def __init__(self, provider, cursor=None, from_cursor=None, to_cursor=None):
        self.provider = provider
        self.cursor = cursor
        self.from_cursor = from_cursor
        self.to_cursor = to_cursor
        self.oauth2_store = OAuth2Store(provider)
        # While we poll a given provider, it may aggregate data from several providers
        # There's no difference yet between a service provider and a pure data provider
        # (these are sets of UUIDs, not str)
        self.provider_uids = set(models.Provider.objects.values_list("pk", flat=True))
        self.device_uids = set(models.Device.objects.values_list("pk", flat=True))

    def poll(self):
        if not self.provider.base_api_url:
            logger.debug("Provider %s has no URL, skipping.", self.provider.name)
            return

        # We request a specific version of the Provider API
        # TODO(hcauwelier) make it mandatory, see SMP-1673
        api_version_raw = self.provider.api_configuration.get(
            "api_version", enums.DEFAULT_PROVIDER_API_VERSION
        )
        # We store the enum key, which cannot be a numeric identifier
        # It doubles as a validity check
        api_version = enums.MDS_VERSIONS[api_version_raw].value

        logger.debug("Polling %s using version %s", self.provider.name, api_version)

        getattr(self, "_poll_status_changes_%s" % api_version_raw)(api_version)

    # TODO(hcauwelier) Should be deleted ASAP
    def _poll_status_changes_v0_2(self, api_version):
        # Start where we left, it's all based on providers sorting by start_time
        # (but we would miss telemetries older than start_time saved after we polled).
        # For those that support it, use the recorded time field or equivalent.
        polling_cursor = self.cursor or self.provider.api_configuration.get(
            "polling_cursor", POLLING_CURSORS.start_time.name
        )
        logger.info(
            f"Starting polling using the field: {polling_cursor}. Current state:\n"
            + (
                (
                    f"\tLast event_time: {str(self.provider.last_event_time_polled)},\n"
                    + f"\tLast recorded: {str(self.provider.last_recorded_polled)}\n"
                    + f"\tLast skip: {self.provider.last_skip_polled}."
                )
                if not self.cursor
                else f"\tCursor value: {self.from_cursor}."
            )
        )

        params = {}

        # Cursor param (raise if typo)
        param_name = {enum.name: enum.value for enum in POLLING_CURSORS}[polling_cursor]

        if polling_cursor == POLLING_CURSORS.total_events.name:
            # Resume from the last line fetched
            params[param_name] = self.from_cursor or self.provider.last_skip_polled
        elif polling_cursor == POLLING_CURSORS.start_recorded.name:
            # Resume from the last "recorded" value
            params[param_name] = utils.to_mds_timestamp(
                self.from_cursor or self.provider.last_recorded_polled
            )
        elif polling_cursor == POLLING_CURSORS.start_time.name:
            # Resume from the last "event_time" value
            last_event_time_polled = self.provider.last_event_time_polled
            if not last_event_time_polled:
                last_event_time_polled = timezone.now() - datetime.timedelta(
                    getattr(settings, "PROVIDER_POLLER_LIMIT", 90)
                )

            # But we now apply a "lag" before actually polling,
            # leaving time for the provider to collect data from its devices
            polling_lag = self.provider.api_configuration.get("provider_polling_lag")
            if polling_lag:
                polling_lag = parse_duration(polling_lag)
                if (timezone.now() - last_event_time_polled) < polling_lag:
                    logger.debug("Still under the polling lag, back to sleep.")
                    return

            params[param_name] = utils.to_mds_timestamp(
                self.from_cursor or last_event_time_polled
            )

        # Provider-specific params to optimise polling
        try:
            params.update(self.provider.api_configuration["status_changes_params"])
        except KeyError:
            pass

        next_url = urllib.parse.urljoin(self.provider.base_api_url, "status_changes")
        if self.provider.api_configuration.get("trailing_slash"):
            next_url += "/"

        if params:
            next_url = "%s?%s" % (next_url, urllib.parse.urlencode(params))

        skip_polled = (
            self.from_cursor
            if self.from_cursor and self.cursor == POLLING_CURSORS.total_events.name
            else None
        )

        # Pagination
        while next_url:
            body = self._get_body(next_url, api_version)
            # Translate older versions of data
            translated_data = translate_v0_2_to_v0_4(body["data"])
            status_changes = translated_data["status_changes"]
            if not status_changes:
                break

            next_url = body.get("links", {}).get("next")

            # A transaction for each "page" of data
            with transaction.atomic():
                # We get the maximum of the recorded and event_types
                # from the status changes
                event_time_polled, recorded_polled = self._process_status_changes(
                    status_changes
                )

                if self.cursor:
                    if self.cursor == POLLING_CURSORS.total_events.name:
                        skip_polled += len(status_changes)
                        if skip_polled >= self.to_cursor:
                            break
                    elif self.cursor == POLLING_CURSORS.start_recorded.name:
                        if recorded_polled >= self.to_cursor:
                            break
                    elif self.cursor == POLLING_CURSORS.start_time.name:
                        if event_time_polled >= self.to_cursor:
                            break
                else:
                    # We get the new skip from the number of status changes
                    skip_polled = (
                        self.provider.last_skip_polled + len(status_changes)
                        if self.provider.last_skip_polled
                        else len(status_changes)
                    )
                    self.provider.last_event_time_polled = event_time_polled
                    self.provider.last_recorded_polled = recorded_polled
                    self.provider.last_skip_polled = skip_polled
                    self.provider.save(
                        update_fields=[
                            "last_event_time_polled",
                            "last_recorded_polled",
                            "last_skip_polled",
                        ]
                    )

                logger.info(
                    f"Polled page using cursor: {polling_cursor}. New state:\n"
                    + f"\tLast event_time: {str(event_time_polled)},\n"
                    + f"\tLast recorded: {str(recorded_polled)}\n"
                    + f"\tLast skip: {skip_polled}."
                )

    def _poll_status_changes_v0_3(self, api_version):
        # Start where we left, it's all based on providers sorting by start_time
        # (but we would miss telemetries older than start_time saved after we polled).
        # For those that support it, use the recorded time field or equivalent.
        polling_cursor = self.cursor or self.provider.api_configuration.get(
            "polling_cursor", POLLING_CURSORS.start_time.name
        )
        logger.info(
            f"Starting polling using the field: {polling_cursor}. Current state:\n"
            + (
                (
                    f"\tLast event_time: {str(self.provider.last_event_time_polled)},\n"
                    + f"\tLast recorded: {str(self.provider.last_recorded_polled)}\n"
                    + f"\tLast skip: {self.provider.last_skip_polled}."
                )
                if not self.cursor
                else f"\tCursor value: {self.from_cursor}."
            )
        )

        params = {}

        # Cursor param (raise if typo)
        param_name = {enum.name: enum.value for enum in POLLING_CURSORS}[polling_cursor]

        if polling_cursor == POLLING_CURSORS.total_events.name:
            # Resume from the last line fetched
            params[param_name] = self.from_cursor or self.provider.last_skip_polled
        elif polling_cursor == POLLING_CURSORS.start_recorded.name:
            # Resume from the last "recorded" value
            params[param_name] = utils.to_mds_timestamp(
                self.from_cursor or self.provider.last_recorded_polled
            )
        elif polling_cursor == POLLING_CURSORS.start_time.name:
            # Resume from the last "event_time" value
            last_event_time_polled = self.provider.last_event_time_polled
            if not last_event_time_polled:
                last_event_time_polled = timezone.now() - datetime.timedelta(
                    getattr(settings, "PROVIDER_POLLER_LIMIT", 90)
                )

            # But we now apply a "lag" before actually polling,
            # leaving time for the provider to collect data from its devices
            polling_lag = self.provider.api_configuration.get("provider_polling_lag")
            if polling_lag:
                polling_lag = parse_duration(polling_lag)
                if (timezone.now() - last_event_time_polled) < polling_lag:
                    logger.debug("Still under the polling lag, back to sleep.")
                    return

            params[param_name] = utils.to_mds_timestamp(
                self.from_cursor or last_event_time_polled
            )

        # Provider-specific params to optimise polling
        try:
            params.update(self.provider.api_configuration["status_changes_params"])
        except KeyError:
            pass

        next_url = urllib.parse.urljoin(self.provider.base_api_url, "status_changes")
        if self.provider.api_configuration.get("trailing_slash"):
            next_url += "/"

        if params:
            next_url = "%s?%s" % (next_url, urllib.parse.urlencode(params))

        skip_polled = (
            self.from_cursor
            if self.from_cursor and self.cursor == POLLING_CURSORS.total_events.name
            else None
        )

        # Pagination
        while next_url:
            body = self._get_body(next_url, api_version)
            # MDS 0.3 is backwards compatible with 0.4
            status_changes = body["data"]["status_changes"]
            if not status_changes:
                break

            next_url = body.get("links", {}).get("next")

            # A transaction for each "page" of data
            with transaction.atomic():
                # We get the maximum of the recorded and event_types
                # from the status changes
                event_time_polled, recorded_polled = self._process_status_changes(
                    status_changes
                )

                if self.cursor:
                    if self.cursor == POLLING_CURSORS.total_events.name:
                        skip_polled += len(status_changes)
                        if skip_polled >= self.to_cursor:
                            break
                    elif self.cursor == POLLING_CURSORS.start_recorded.name:
                        if recorded_polled >= self.to_cursor:
                            break
                    elif self.cursor == POLLING_CURSORS.start_time.name:
                        if event_time_polled >= self.to_cursor:
                            break
                else:
                    # We get the new skip from the number of status changes
                    skip_polled = (
                        self.provider.last_skip_polled + len(status_changes)
                        if self.provider.last_skip_polled
                        else len(status_changes)
                    )
                    self.provider.last_event_time_polled = event_time_polled
                    self.provider.last_recorded_polled = recorded_polled
                    self.provider.last_skip_polled = skip_polled
                    self.provider.save(
                        update_fields=[
                            "last_event_time_polled",
                            "last_recorded_polled",
                            "last_skip_polled",
                        ]
                    )

                logger.info(
                    f"Polled page using cursor: {polling_cursor}. New state:\n"
                    + f"\tLast event_time: {str(event_time_polled)},\n"
                    + f"\tLast recorded: {str(recorded_polled)}\n"
                    + f"\tLast skip: {skip_polled}."
                )

    def _poll_status_changes_v0_4(self, api_version):
        # Start where we left, it's all based on providers sorting by start_time
        # (but we would miss telemetries older than start_time saved after we polled).
        # For those that support it, use the recorded time field or equivalent.
        polling_cursor = self.cursor or self.provider.api_configuration.get(
            "polling_cursor", POLLING_CURSORS.start_time.name
        )
        if polling_cursor != POLLING_CURSORS.start_time.name:
            raise ValueError('Only "start_time" cursor is supported in MDS 0.4+')
        logger.info(
            f"Starting polling using the field: {polling_cursor}. Current state:\n"
            + (
                (f"\tLast event_time: {str(self.provider.last_event_time_polled)}.")
                if not self.cursor
                else f"\tCursor value: {self.from_cursor}."
            )
        )

        params = {}

        # Resume from the last "event_time" value
        last_event_time_polled = self.provider.last_event_time_polled
        if not last_event_time_polled:
            last_event_time_polled = timezone.now() - datetime.timedelta(
                getattr(settings, "PROVIDER_POLLER_LIMIT", 90)
            )

        # But we now apply a "lag" before actually polling,
        # leaving time for the provider to collect data from its devices
        polling_lag = self.provider.api_configuration.get("provider_polling_lag")
        if polling_lag:
            polling_lag = parse_duration(polling_lag)
            if (timezone.now() - last_event_time_polled) < polling_lag:
                logger.info("Still under the polling lag, back to sleep.")
                return

        # The MDS 0.4 Provider API got two endpoints:
        # - /events for the real-time or so data (formerly /status_changes)
        #   but limited to two weeks of history
        # - /status_changes for all the data except the current hour
        # If we're catching up far in time, begin by polling /status_changes
        realtime_threshold = parse_duration(  # Default is 9 days
            self.provider.api_configuration.get("realtime_threshold", "P9D")
        )
        if (timezone.now() - last_event_time_polled) > realtime_threshold:
            # We have to query the archived status changes with another format
            logger.info("last_event_time_polled is too old, asking archives")
            params["event_time"] = last_event_time_polled.isoformat()[
                : len("YYYY-MM-DDTHH")
            ]
        else:
            # Query the real-time endpoint as usual
            params["start_time"] = utils.to_mds_timestamp(
                self.from_cursor or last_event_time_polled
            )

        # Provider-specific params to optimise polling
        try:
            params.update(self.provider.api_configuration["status_changes_params"])
        except KeyError:
            pass

        endpoint = "events"  # The new name for the real-time events endpoint
        if "event_time" in params:
            # We asked the archived status changes instead
            endpoint = "status_changes"
        next_url = urllib.parse.urljoin(self.provider.base_api_url, endpoint)
        if self.provider.api_configuration.get("trailing_slash"):
            next_url += "/"

        if params:
            next_url = "%s?%s" % (next_url, urllib.parse.urlencode(params))

        # Pagination
        while next_url:
            body = self._get_body(next_url, api_version)
            # No translation needed as long as 0.4 is the latest version
            status_changes = body["data"]["status_changes"]
            next_url = body.get("links", {}).get("next")

            # A transaction for each "page" of data
            with transaction.atomic():
                if status_changes:
                    # We get the maximum values from the status changes
                    event_time_polled, _ = self._process_status_changes(status_changes)
                else:
                    if endpoint == "status_changes":
                        # this hour frame of archives didn't contain results
                        event_time_polled = last_event_time_polled + datetime.timedelta(
                            hours=1
                        )
                    else:
                        break

                if self.cursor:
                    if self.cursor == POLLING_CURSORS.start_time.name:
                        if event_time_polled >= self.to_cursor:
                            break
                else:
                    self.provider.last_event_time_polled = event_time_polled
                    self.provider.save(update_fields=["last_event_time_polled"])

                logger.info(
                    f"Polled page using cursor: {polling_cursor}. New state:\n"
                    + f"\tLast event_time: {str(event_time_polled)}."
                )

    @retry(stop_max_attempt_number=2)
    def _get_body(self, url, api_version):
        authentication_type = self.provider.api_authentication.get("type")
        if authentication_type in (None, "", "none"):
            client = requests
        elif authentication_type == "oauth2":
            client = self.oauth2_store.get_client()
        else:
            raise NotImplementedError

        headers = {}

        # TODO(hcauwelier) just fork the code for MDS 0.4
        if api_version == "0.4":
            # We only accept the version we expect from that provider
            # And it should reject it when it bumps to the new version
            headers["Accept"] = "%s;version=%s" % (MDS_CONTENT_TYPE, api_version)
        else:
            # TODO(hcauwelier) some server implementations are flawed
            # leave the standard content type for now
            headers["Accept"] = "application/json,%s;version=%s" % (
                MDS_CONTENT_TYPE,
                api_version,
            )

        logger.debug("Polling provider on URL %s with headers %s", url, headers)
        response = client.get(url, timeout=30, headers=headers)
        # Token may be expired sooner than expected, retry in one minute
        if response.status_code in (401, 403):
            self.oauth2_store.flush_token()
        response.raise_for_status()

        # Servers should send what version of the API was served
        # but we rather trust the versioning in the body
        content_type = response.headers.get(
            "Content-Type", "application/json; charset=UTF-8"
        )
        if content_type.startswith("application/json"):
            received = "unspecified"
        else:
            # Let it raise if malformed
            params = content_type.split(";")
            content_type = params.pop(0)
            assert content_type == MDS_CONTENT_TYPE
            params = dict(param.split("=") for param in params)
            received = params.get("version", "unspecified")
        if received != api_version:
            logger.warning(
                "Provider %s didn't respond in version %s but %s",
                self.provider.name,
                api_version,
                received,
            )

        body = response.json()
        return body

    def _process_status_changes(self, status_changes):
        logger.debug("Processing...")

        # accept timestamp as a string instead of an integer
        status_changes = self._validate_event_times(status_changes)
        if not status_changes:
            # Data so bad there is no or nothing but invalid event times
            logger.exception(
                "No valid event_time found in status_changes series: %s", status_changes
            )
            # How can we prevent from asking them again next time?
            if (
                self.provider.last_event_time_polled
                and self.provider.last_recorded_polled
            ):
                ms = datetime.timedelta(milliseconds=1)
                return (
                    self.provider.last_event_time_polled + ms,
                    self.provider.last_recorded_polled + ms,
                )

            # The provider really doesn't help!
            return timezone.now(), timezone.now()

        last_event_time_polled = utils.from_mds_timestamp(
            max(status_change["event_time"] for status_change in status_changes)
        )
        last_recorded_polled = utils.from_mds_timestamp(
            max(status_change.get("recorded") or 0 for status_change in status_changes)
        )

        status_changes = self._validate_status_changes(status_changes)
        if not status_changes:
            # None were valid, we won't ask that series again
            # (provided status changes are ordered by event_time ascending)
            return last_event_time_polled, last_recorded_polled

        self._create_missing_providers(status_changes)
        self._create_missing_devices(status_changes)
        self._create_event_records(status_changes)

        return last_event_time_polled, last_recorded_polled

    def _validate_event_times(self, status_changes):
        """I need this one done before validating the rest of the data."""
        validated_status_changes = []

        for status_change in status_changes:
            try:
                status_change["event_time"] = int(status_change["event_time"])
            except (KeyError, ValueError):
                logger.warning(
                    "status_change %s has no valid event_time", status_change
                )
                continue
            validated_status_changes.append(status_change)

        return validated_status_changes

    def _validate_status_changes(self, status_changes):
        """Some preliminary checks/addenda"""
        validated_status_changes = []

        for status_change in status_changes:
            status_change["provider_id"] = uuid.UUID(status_change["provider_id"])
            status_change["device_id"] = device_id = uuid.UUID(
                status_change["device_id"]
            )

            # The list of event types and even the naming don't match between
            # the provider and agency APIs, so translate one to the other
            try:
                event_type_reason = status_change["event_type_reason"]
            except KeyError:  # Spec violation!
                logger.warning("Device %s has no event_type_reason", device_id)
                # Ignore just that status change to avoid rejecting the whole batch
                continue
            try:
                event = PROVIDER_REASON_TO_AGENCY_EVENT[event_type_reason]
                agency_event_type, agency_event_type_reason = (
                    event if len(event) == 2 else event + (None,)
                )
            except KeyError:  # Spec violation!
                logger.warning(
                    'Device %s has unknown "%s" event_type_reason',
                    device_id,
                    event_type_reason,
                )
                # Ignore just that status change to avoid rejecting the whole batch
                continue
            status_change["agency_event_type"] = agency_event_type
            status_change["agency_event_type_reason"] = agency_event_type_reason

            event_location = status_change["event_location"]
            if event_location:
                assert event_location["geometry"]["type"] == "Point"
                # Some providers may get the (lng, lat) order wrong
                if self.provider.api_configuration.get("swap_lng_lat"):
                    lat, lng = event_location["geometry"]["coordinates"][:2]
                    event_location["geometry"]["coordinates"][0] = lng
                    event_location["geometry"]["coordinates"][1] = lat
            else:  # Spec violation!
                logger.warning("Device %s has no event_location", device_id)
                # This time, accept a status change with no location

            validated_status_changes.append(status_change)

        return validated_status_changes

    def _create_missing_providers(self, status_changes):
        """Make sure all providers mentioned exist"""

        with_missing_providers = [
            status_change
            for status_change in status_changes
            if status_change["provider_id"] not in self.provider_uids
        ]

        if with_missing_providers:
            db_helpers.upsert_providers(
                (
                    _create_provider(status_change)
                    for status_change in with_missing_providers
                )
            )

            providers_added = [
                status_change["provider_id"] for status_change in with_missing_providers
            ]
            self.provider_uids.update(providers_added)

            logger.info(
                "Providers created: %s", ", ".join(str(uid) for uid in providers_added)
            )

    def _create_missing_devices(self, status_changes):
        """Make sure all devices mentioned exist"""

        with_missing_devices = [
            status_change
            for status_change in status_changes
            if status_change["device_id"] not in self.device_uids
        ]

        if with_missing_devices:
            db_helpers.upsert_devices(
                (
                    _create_device(status_change)
                    for status_change in with_missing_devices
                )
            )
            if getattr(settings, "POLLER_CREATE_REGISTER_EVENTS", False):
                # Create fake register events to simulate device registration
                db_helpers.upsert_event_records(
                    (
                        _create_register_event_record(status_change)
                        for status_change in with_missing_devices
                    ),
                    source=enums.EVENT_SOURCE.provider_api.name,
                )

            devices_added = [
                status_change["device_id"] for status_change in with_missing_devices
            ]
            self.device_uids.update(devices_added)

            logger.info(
                "Devices created: %s", ", ".join(str(uid) for uid in devices_added)
            )

    def _create_event_records(self, status_changes):
        """Now record the... records"""
        db_helpers.upsert_event_records(
            (_create_event_record(status_change) for status_change in status_changes),
            enums.EVENT_SOURCE.provider_api.name,
            # Timestamps are unique per device, ignore duplicates
            # Events already pushed by the provider will always have precedence
            on_conflict_update=False,
        )


def _create_provider(status_change):
    provider_id = status_change["provider_id"]
    try:
        name = status_change["provider_name"]
    except KeyError:  # Spec violation!
        logger.warning("Provider %s has no name", provider_id)
        name = ""

    return models.Provider(id=provider_id, name=name)


def _create_device(status_change):
    device_id = status_change["device_id"]
    identification_number = status_change["vehicle_id"]
    if not identification_number:  # Spec violation!
        logger.warning("Device %s has no identification number", device_id)
        identification_number = "test-%s" % str(device_id).split("-", 1)

    return models.Device(
        id=device_id,
        # Don't assume the device received belongs to the provider requested
        # The LA sandbox contains data for several providers
        provider_id=status_change["provider_id"],
        identification_number=identification_number,
        category=status_change["vehicle_type"],
        propulsion=status_change["propulsion_type"],
    )


def _create_event_record(status_change):
    properties = {"trip_id": status_change.get("associated_trip")}

    event_location = status_change["event_location"]
    if event_location:
        # GeoJSON Point Feature
        try:
            longitude, latitude, altitude = event_location["geometry"]["coordinates"]
        except ValueError:
            longitude, latitude = event_location["geometry"]["coordinates"]
            altitude = None
        point = geos.Point(longitude, latitude, altitude, srid=4326)
        properties["telemetry"] = {
            "timestamp": event_location["properties"]["timestamp"],
            "gps": {"lng": longitude, "lat": latitude},
            # No coordinates, no battery charge saved
            "battery_pct": status_change.get("battery_pct"),
        }
        if altitude:
            properties["telemetry"]["gps"]["altitude"] = altitude
    else:  # Spec violation!
        point = None

    publication_time = None
    if status_change.get("publication_time"):
        publication_time = utils.from_mds_timestamp(status_change["publication_time"])
    # "Aggregation" providers store a recorded field, and we want to keep the same value
    # until we get the publication_time everywhere
    # Vendor only, 0.3 only, will disappear
    recorded = None
    if status_change.get("recorded"):
        recorded = utils.from_mds_timestamp(status_change["recorded"])
    if publication_time and recorded:
        difference = abs(publication_time - recorded)
        if difference > datetime.timedelta(minutes=10):
            logger.warning("publication_time and recorded differ by %s", difference)

    return models.EventRecord(
        device_id=status_change["device_id"],
        timestamp=utils.from_mds_timestamp(status_change["event_time"]),
        point=point,
        event_type=status_change["agency_event_type"],
        event_type_reason=status_change["agency_event_type_reason"],
        properties=properties,
        publication_time=publication_time or recorded,
    )


def _create_register_event_record(status_change):
    """
    As the goal of the poller is to catch up with the history of a provider,
    simulate the registration of a device with a fake register event.

    This fake event is flagged in its properties to tell it apart from real ones.

    Should the device be unregistered and registered again according to the specs,
    don't delete the fake events in the past.
    """
    return models.EventRecord(
        device_id=status_change["device_id"],
        # Another event for the same device with the same timestamp will be rejected
        timestamp=utils.from_mds_timestamp(status_change["event_time"])
        - datetime.timedelta(milliseconds=1),
        event_type=enums.EVENT_TYPE.register.name,
        properties={"created_on_register": True},
        source=enums.EVENT_SOURCE.provider_api.name,
    )



# Generated by Django 2.2.5 on 2019-12-20 11:44

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):
    deploy_phase = "pre_deploy"

    dependencies = [("mds", "0001_squashed_0059_post_policy_config_encoder")]

    operations = [
        migrations.AddField(
            model_name="compliance",
            name="extra",
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True),
        )
    ]



from typing import Set, Optional, List, Dict

import jwt
from django.utils.translation import ugettext as _
from rest_framework import exceptions

from .auth_means import BaseAuthMean
from .jwt_decode import jwt_multi_decode

from mds.authent.public_api import get_revocation_list


class RemoteUser:
    """
    Inspired by django.contrib.auth.models.AnonymousUser
    """

    _id: str
    _scopes: Set[str] = set()
    _aggregator_for: Set[str] = set()
    _provider_id: Optional[str] = None
    """If provided, restrict access to data owned by given provider"""

    def __init__(self, sub, scopes, aggregator_for, provider_id):
        self._id = sub
        self._scopes = scopes
        self._aggregator_for = aggregator_for
        self._provider_id = provider_id

    def __str__(self):
        return "RemoteUser {}".format(self._id)

    @property
    def is_anonymous(self):
        return False

    @property
    def is_authenticated(self):
        return True

    @property
    def is_staff(self):
        return False

    @property
    def scopes(self):
        return self._scopes

    @property
    def aggregator_for(self):
        return self._aggregator_for

    @property
    def provider_id(self):
        return self._provider_id


def authenticate(auth_means: List[BaseAuthMean], encoded_jwt: str) -> RemoteUser:
    # Step 1: check and decode token
    try:
        payload, introspect_url = jwt_multi_decode(auth_means, encoded_jwt)
    except jwt.ExpiredSignature:
        msg = _("Signature has expired.")
        raise exceptions.AuthenticationFailed(msg)
    except jwt.DecodeError:
        msg = _("Error decoding signature.")
        raise exceptions.AuthenticationFailed(msg)
    except jwt.InvalidTokenError:
        raise exceptions.AuthenticationFailed()

    # Step 2: optionally check token validity
    if payload.get("jti") in get_revocation_list():
        raise exceptions.AuthenticationFailed(_("Expired or revoked token"))

    # Step 3: build user
    user = build_user(payload)

    return user


def build_user(payload: Dict) -> RemoteUser:
    """
    Returns a user from the JWT payload

    Fields looked-for in the payload:
        * sub (required): identifier for the owner of the token. Could be a human user,
          a provider's server, ...
        * jti (required): identifier for the JWT (make it possible to blacklist
          the token)
        * scope (required): space-delimited permissions
        * app_owner (optional, required for provider): asked for by
          https://github.com/CityOfLosAngeles/mobility-data-specification/tree/dev/agency#authorization
    """
    required_fields = {"sub", "jti", "scope"}
    missing_fields = required_fields - payload.keys()

    if missing_fields:
        msg = _("Invalid payload, missing fields: %(fields)s")
        raise exceptions.AuthenticationFailed(
            msg % {"fields": ", ".join(missing_fields)}
        )

    # See https://tools.ietf.org/html/rfc6749#section-3.3
    scopes = set(payload["scope"].split(" "))
    aggregator_for = set(payload.get("aggregator_for", "").split(" "))

    return RemoteUser(
        payload["sub"], scopes, aggregator_for, payload.get("app_owner", None)
    )



import datetime

import oauthlib.oauth2
import oauthlib.oauth2.rfc6749.tokens
import oauth2_provider.models
import oauth2_provider.oauth2_validators
from oauth2_provider.scopes import BaseScopes
from oauth2_provider.settings import oauth2_settings


from . import generators


def signed_token_generator(request):
    token_duration = datetime.timedelta(
        seconds=oauth2_settings.ACCESS_TOKEN_EXPIRE_SECONDS
    )
    user = getattr(request, "user", None)
    token, payload = generators.generate_jwt_with_payload(
        request.client, token_duration=token_duration, user=user
    )
    # set claims on the request
    request.claims = payload
    return token


class Server(oauthlib.oauth2.Server):
    """Just swap the default token generator to signed_tokens."""

    def __init__(self, *args, **kwargs):
        if not kwargs.get("token_generator"):
            kwargs["token_generator"] = signed_token_generator
            kwargs[
                "refresh_token_generator"
            ] = oauthlib.oauth2.rfc6749.tokens.random_token_generator
        super().__init__(*args, **kwargs)


class OAuth2Validator(oauth2_provider.oauth2_validators.OAuth2Validator):
    def _create_access_token(self, expires, request, token, source_refresh_token=None):
        """Saves the token jti in the database."""
        access_token = oauth2_provider.models.get_access_token_model()(
            user=request.user,
            scope=token["scope"],
            expires=expires,
            token=token["access_token"],
            application=request.client,
            source_refresh_token=source_refresh_token,
            jti=request.claims["jti"],
        )
        access_token.save()
        return access_token


class AppScopes(BaseScopes):
    def get_all_scopes(self):
        Application = oauth2_provider.models.get_application_model()
        return {
            k: k
            for k in set(
                s
                for scopes in Application.objects.values_list("scopes", flat=True)
                for s in scopes
            )
        }

    def get_available_scopes(self, application=None, request=None, *args, **kwargs):
        return application.scopes

    def get_default_scopes(self, application=None, request=None, *args, **kwargs):
        return self.get_available_scopes(application, request, *args, **kwargs)



"""A few helpers around Swagger."""
from django_filters import rest_framework as filters
from drf_yasg.inspectors import SwaggerAutoSchema
from drf_yasg.inspectors.base import FilterInspector, NotHandled
import drf_yasg.inspectors.base
import drf_yasg.openapi
import drf_yasg.utils


def _call_view_method(
    view, method_name, fallback_attr=None, default=None, args=None, kwargs=None
):
    """Override of drf_yasg.inspectors.base.call_view_method to allow passing args."""
    if hasattr(view, method_name):
        try:
            view_method, is_callabale = drf_yasg.inspectors.base.is_callable_method(
                view, method_name
            )
            if is_callabale:
                args = args or []
                kwargs = kwargs or {}
                return view_method(*args, **kwargs)
        except Exception:  # noqa
            drf_yasg.inspectors.base.logger.warning(
                "view's %s raised exception during schema generation; use "
                "`getattr(self, 'swagger_fake_view', False)` "
                "to detect and short-circuit this",
                type(view).__name__,
                exc_info=True,
            )

    if fallback_attr and hasattr(view, fallback_attr):
        return getattr(view, fallback_attr)

    return default


class CustomSwaggerAutoSchema(SwaggerAutoSchema):
    """Customer Schema generator that understands Request and Response serializers.

    By default DRF has no concept of Request serializer VS Response serializer.
    A view only defines one serializer which could be limiting.
    This Schema generator, in combination with MultiSerializerViewSetMixin bypasses
    this limitation.
    """

    def get_view_serializer(self, request_or_response):
        """Return the serializer as defined by the view's ``get_serializer()`` method.

        :return: the view's ``Serializer``
        :rtype: rest_framework.serializers.Serializer
        """
        return _call_view_method(
            self.view,
            "get_serializer",
            kwargs={"context": {"request_or_response": request_or_response}},
        )

    def get_request_serializer(self):
        """Return the request serializer for this endpoint.

        :return: the request serializer, or one of :class:`.Schema`,
                 :class:`.SchemaRef`, ``None``
        :rtype: rest_framework.serializers.Serializer
        """
        body_override = self._get_request_body_override()

        if body_override is None and self.method in self.implicit_body_methods:
            return self.get_view_serializer("request")

        if body_override is drf_yasg.utils.no_body:
            return None

        return body_override

    def get_default_response_serializer(self):
        """Return the default response serializer for this endpoint.

        This is derived from either the ``request_body`` override
        or the request serializer (:meth:`.get_view_serializer`).

        :return: response serializer, :class:`.Schema`, :class:`.SchemaRef`, ``None``
        """
        body_override = self._get_request_body_override()
        if body_override and body_override is not drf_yasg.utils.no_body:
            return body_override

        return self.get_view_serializer("response")


class CustomFilterInspector(FilterInspector):
    def _convert_range_field_to_parameters(self, field_name, field):
        """ Generate two separate query parameters for range filters """
        if isinstance(field, filters.NumberFilter):
            type_ = drf_yasg.openapi.TYPE_NUMBER
        else:
            type_ = drf_yasg.openapi.TYPE_STRING
        common_params = dict(
            in_=drf_yasg.openapi.IN_QUERY, required=field.extra["required"], type=type_
        )
        return (
            drf_yasg.openapi.Parameter(
                name=f"{field_name}_{field.field_class.widget.suffixes[0]}",
                description=f'{field.extra.get("help_text", "range")} (lower bound)',
                **common_params,
            ),
            drf_yasg.openapi.Parameter(
                name=f"{field_name}_{field.field_class.widget.suffixes[1]}",
                description=f'{field.extra.get("help_text", "range")} (upper bound)',
                **common_params,
            ),
        )

    def _convert_field_to_parameters(self, field_name, field):
        extra_param = {}
        if isinstance(field, filters.RangeFilter):
            return self._convert_range_field_to_parameters(field_name, field)

        if isinstance(field, filters.MultipleChoiceFilter):
            type_ = drf_yasg.openapi.TYPE_ARRAY
            extra_param["items"] = drf_yasg.openapi.Items(
                type=drf_yasg.openapi.TYPE_STRING,
                enum=[c for c, _ in field.extra["choices"]],
            )
            extra_param["collection_format"] = "multi"
        else:
            if isinstance(field, filters.ModelChoiceFilter):
                type_ = subtype = drf_yasg.openapi.TYPE_STRING
                # extra_param?
            elif isinstance(field, filters.ChoiceFilter):
                type_ = subtype = drf_yasg.openapi.TYPE_STRING
                extra_param["enum"] = [c for c, _ in field.extra["choices"]]
            elif isinstance(field, filters.NumberFilter):
                type_ = subtype = drf_yasg.openapi.TYPE_NUMBER
            else:
                type_ = subtype = drf_yasg.openapi.TYPE_STRING
            if field.lookup_expr == "in":
                type_ = drf_yasg.openapi.TYPE_ARRAY
                extra_param["items"] = drf_yasg.openapi.Items(
                    type=subtype, enum=extra_param.pop("enum", None)
                )
        return (
            drf_yasg.openapi.Parameter(
                name=field_name,
                in_=drf_yasg.openapi.IN_QUERY,
                required=field.extra["required"],
                description=field.extra.get("help_text", ""),
                type=type_,
                **extra_param,
            ),
        )

    def get_filter_parameters(self, filter_backend):
        if not hasattr(filter_backend, "get_filterset_class"):
            return NotHandled

        queryset = self.view.get_queryset()
        filterset_class = filter_backend.get_filterset_class(self.view, queryset)

        parameters = []
        if filterset_class:
            for field_name, field in filterset_class.base_filters.items():
                parameters += self._convert_field_to_parameters(field_name, field)
        return parameters



"""
Database description
"""
import json
import uuid

from django import forms
from django.contrib.gis.db import models as gis_models
from django.contrib.postgres import fields as pg_fields
from django.contrib.postgres import functions as pg_functions
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Count, Prefetch, Q, Index
from django.utils import timezone
from django.core.serializers.json import DjangoJSONEncoder

from rest_framework.utils import encoders

from . import enums


class UnboundedCharField(models.TextField):
    """Unlimited text, on a single line.

    Shows an ``<input type="text">`` in HTML but is stored as a TEXT
    column in Postgres (like ``TextField``).

    Like the standard :class:`~django.db.models.fields.CharField` widget,
    a ``select`` widget is automatically used if the field defines ``choices``.
    """

    def formfield(self, **kwargs):
        kwargs["widget"] = None if self.choices else forms.TextInput
        return super().formfield(**kwargs)


def agency_api_authentication_default():
    """Default value for the ``agency_api_authentication`` field.

    Other keys are specific to each type:
    - token: "header" and "token"
    - oauth2: "client_id" and "client_secret"
    """
    return {"type": "none"}


def agency_api_configuration_default():
    """Default value for the ``agency_api_configuration`` field."""
    return {}


def provider_api_authentication_default():
    """Default value for the API authentication field of a provider.

    Other keys are specific to each type:
    - token: "header" and "token"
    - oauth2: "client_id" and "client_secret"
    """
    return {"type": "none"}


def provider_api_configuration_default():
    """Default value for the API configuration field of a provider.

    Some provider implementations can be picky, configure these special cases here.
    """
    return {"trailing_slash": False}  # Some providers endpoint won't reply without it


def short_uuid4(uid):
    """When seeing a glimpse of the UID is enough."""
    return str(uid)[:8]  # Basically splitting on the first dash


class ProviderQuerySet(models.QuerySet):
    def with_device_categories(self, **kwargs):
        # A single dict would be simpler but I don't know how to make this in SQL
        # (ERROR:  subquery must return only one column)
        device_categories = {}
        for device_category in enums.DEVICE_CATEGORY:
            device_categories["device_category_%s" % device_category.name] = Count(
                "devices", filter=Q(devices__category=device_category.name)
            )
        return self.filter(**kwargs).annotate(**device_categories)

    def operators(self, **kwargs):
        return self.filter(operator=True, **kwargs)


def colors_validator(colors):
    fields = {"primary", "secondary"}
    invalid_keys = set(colors.keys()) - fields
    if invalid_keys:
        raise ValidationError(
            f"{invalid_keys} is not an accepted field. Field must be in {fields}"
        )


class Provider(models.Model):

    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    name = UnboundedCharField(blank=True, default="")
    logo_b64 = UnboundedCharField(null=True, blank=True, default=None)
    operator = models.BooleanField(default=True)

    # The following fields are for us pulling data from the provider
    base_api_url = UnboundedCharField(
        default="", blank=True, verbose_name="Base Provider API URL"
    )
    oauth2_url = UnboundedCharField(
        default="", blank=True, verbose_name="OAuth2 URL (if different)"
    )
    api_authentication = pg_fields.JSONField(
        default=provider_api_authentication_default,
        blank=True,
        verbose_name="Provider API Authentication",
    )
    api_configuration = pg_fields.JSONField(
        default=provider_api_configuration_default,
        blank=True,
        verbose_name="Provider API Configuration",
    )
    # We may poll a provider, e.g. LADOT sandbox that replies for many providers
    # but has no device itself.
    # So we cannot just rely on checking the latest event record saved.
    last_event_time_polled = models.DateTimeField(
        blank=True, null=True, verbose_name="Last event time polled (start_time)"
    )
    last_recorded_polled = models.DateTimeField(
        blank=True, null=True, verbose_name="Last recorded polled (start_recorded)"
    )
    # Some providers accept `skip` as a query param
    last_skip_polled = models.PositiveIntegerField(
        blank=True, null=True, verbose_name="Last skip polled (total_events)"
    )

    # The following fields are for the provider pushing data to us
    agency_api_authentication = pg_fields.JSONField(
        default=agency_api_authentication_default,
        blank=True,
        verbose_name="Agency API Authentication",
    )
    agency_api_configuration = pg_fields.JSONField(
        default=agency_api_configuration_default,
        blank=True,
        verbose_name="Agency API Configuration",
    )
    colors = pg_fields.JSONField(
        default=dict,
        blank=True,
        verbose_name="colors for distinguishing provider icons",
        validators=[colors_validator],
    )

    objects = ProviderQuerySet.as_manager()

    def __str__(self):
        return "{} ({})".format(self.name or "Provider object", short_uuid4(self.id))

    @property
    def device_categories(self):
        # This will fail if you didn't call objects.with_device_categories()
        device_categories = {}
        for category in enums.DEVICE_CATEGORY:
            device_categories[category.name] = getattr(
                self, "device_category_%s" % category.name
            )
        return device_categories


class DeviceQueryset(models.QuerySet):
    def with_latest_events(self):
        prefetched_events = Prefetch(
            "event_records",
            # Excluding telemetry because MDS Agency separates event from telemetry
            # and the "latest_event" does not count telemetries as events
            queryset=EventRecord.objects.exclude(event_type="telemetry").order_by(
                "-timestamp"
            ),
            to_attr="_latest_events",
        )
        # Here, we can't limit the query set so, in order to limit,
        # we have to use a latest_event property (see the property in the model)
        prefetch_related = self.prefetch_related(prefetched_events)
        return prefetch_related


class Device(models.Model):
    """A device
    """

    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    provider = models.ForeignKey(
        Provider, related_name="devices", on_delete=models.CASCADE
    )
    registration_date = models.DateTimeField(default=timezone.now)
    identification_number = UnboundedCharField()

    category = UnboundedCharField(choices=enums.choices(enums.DEVICE_CATEGORY))
    model = UnboundedCharField(blank=True, default="")
    propulsion = pg_fields.ArrayField(
        UnboundedCharField(choices=enums.choices(enums.DEVICE_PROPULSION))
    )
    year_manufactured = models.IntegerField(blank=True, null=True)
    manufacturer = UnboundedCharField(blank=True, default="")
    # Remember to update this field within raw queries!
    saved_at = models.DateTimeField(db_index=True, auto_now=True)

    # denormalized fields - the source of truth is in the EventRecord table.
    # /!\ These fields are for internal usage and may disappear anytime
    dn_battery_pct = models.FloatField(blank=True, null=True)
    dn_gps_point = gis_models.PointField(blank=True, null=True)
    dn_gps_timestamp = models.DateTimeField(blank=True, null=True)
    dn_status = UnboundedCharField(
        choices=enums.choices(enums.DEVICE_STATUS),
        default=enums.DEVICE_STATUS.unknown.name,
    )

    objects = DeviceQueryset.as_manager()

    def __str__(self):
        return "{} {} ({})".format(
            self.get_category_display(),
            self.identification_number,
            short_uuid4(self.id),
        )

    @property
    def latest_event(self):
        if hasattr(self, "_latest_events"):
            # don't do a query in this case, the telemetry was prefetched.
            return self._latest_events[0] if self._latest_events else None
        latest_events = (
            EventRecord.objects.filter(device_id=self.id)
            .exclude(event_type="telemetry")
            .order_by("-timestamp")
        )
        return latest_events.first()

    @property
    def gps_point_as_geojson(self):
        """Represent the GPS point as GeoJSON"""
        if not self.dn_gps_point:
            return None
        return json.loads(self.dn_gps_point.geojson)


class EventRecord(models.Model):
    timestamp = models.DateTimeField(db_index=True)
    point = gis_models.PointField(blank=True, null=True)
    # For synchronisation purposes (polling on this field)
    saved_at = models.DateTimeField(default=pg_functions.TransactionNow, db_index=True)
    # Date/time that event became available through the status changes endpoint
    # (used to be named "first_saved_at", but we won't want the cost of a migration)
    publication_time = models.DateTimeField(
        blank=True, null=True, db_column="first_saved_at"
    )
    source = UnboundedCharField(
        choices=enums.choices(enums.EVENT_SOURCE),
        default=enums.EVENT_SOURCE.agency_api.name,
    )
    device = models.ForeignKey(
        Device, related_name="event_records", on_delete=models.CASCADE
    )
    event_type = UnboundedCharField(
        choices=enums.choices(enums.EVENT_TYPE) + enums.choices(enums.OLD_EVENT_TYPE)
    )
    event_type_reason = UnboundedCharField(
        choices=enums.choices(enums.EVENT_TYPE_REASON), blank=True, null=True
    )

    # JSON fields:
    # {
    #   "telemetry": see factories.EventRecord for an exhaustive example
    # }
    properties = pg_fields.JSONField(default=dict, encoder=encoders.JSONEncoder)

    class Meta:
        unique_together = [("device", "timestamp")]
        # TODO(hcauwelier) only ~2% of the events are useful to display
        # Speed up searching among them for a given device
        indexes = [
            Index(
                fields=["device"],
                name="device_mds_events_partial",
                condition=~Q(event_type="telemetry"),
            )
        ]

    @property
    def point_as_geojson(self):
        """Represent the GPS point as GeoJSON"""
        if not self.point:
            return None
        return json.loads(self.point.geojson)

    @property
    def updated_status(self):
        return enums.EVENT_TYPE_TO_DEVICE_STATUS.get(
            self.event_type, enums.DEVICE_STATUS.unknown.name
        )

    # TODO(hcauwelier) backwards compat, to remove
    @property
    def first_saved_at(self):
        return self.publication_time


class Polygon(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    label = UnboundedCharField(default="", blank=True, db_index=True)
    geom = gis_models.MultiPolygonField()
    properties = pg_fields.JSONField(default=dict, encoder=encoders.JSONEncoder)

    def __str__(self):
        return "{} ({})".format(self.label or "Polygon object", short_uuid4(self.id))


class Area(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    label = UnboundedCharField(default="", blank=True, db_index=True)
    polygons = models.ManyToManyField(Polygon, blank=True, related_name="areas")
    providers = models.ManyToManyField(Provider, blank=True, related_name="areas")
    color = UnboundedCharField(default="#FFFFFF", help_text="hexa representation")
    area_type = UnboundedCharField(
        choices=enums.choices(enums.AREA_TYPE),
        default=enums.AREA_TYPE.unrestricted.name,
    )

    def __str__(self):
        return "{} ({})".format(self.label or "Area object", short_uuid4(self.id))


class PolicyQueryset(models.QuerySet):
    def active(self, at=None, **kwargs):
        if not at:
            at = timezone.now()
        return self.filter(
            Q(end_date__gt=at) | Q(end_date__isnull=True),
            start_date__lte=at,
            # Whether to consider retroactive policies is up to the policy engine
            published_date__isnull=False,
            prev_policies__isnull=True,
            **kwargs,
        )


class Policy(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    name = UnboundedCharField()
    description = UnboundedCharField(default="", blank=True)
    providers = models.ManyToManyField(Provider, blank=True)
    start_date = models.DateTimeField()
    end_date = models.DateTimeField(blank=True, null=True)
    published_date = models.DateTimeField(blank=True, null=True)
    prev_policies = models.ManyToManyField("self", blank=True)
    rules = pg_fields.JSONField(default=list, encoder=DjangoJSONEncoder)
    config = pg_fields.JSONField(default=dict, encoder=DjangoJSONEncoder)
    # Copy of the geographies referenced when the policy is published
    # Only these geographies are admissible when computing the policies
    # So that the originals can be modified without affecting the ongoing policies
    geographies = pg_fields.JSONField(
        default=dict, encoder=DjangoJSONEncoder, blank=True, null=True
    )

    objects = PolicyQueryset.as_manager()

    class Meta:
        verbose_name_plural = "policies"

    def __str__(self):
        return f"{self.name} ({short_uuid4(self.id)})"

    def save(self, *args, **kwargs):
        # Denormalize policy kind with first rule type
        if self.rules:
            self.config.update({"rule_type": self.rules[0]["rule_type"]})
        super(Policy, self).save(*args, **kwargs)

    @property
    def kind(self):
        return self.config["rule_type"]

    def publish(self, at=None):
        """Do the operations required for the policy to be published."""
        if not self.rules:
            raise ValidationError(f"Rules are required on published policies.")

        # "Freeze" the areas' geography in the policy
        # Before publication, rules use the UID of actual areas
        # Dumped geographies use a different UID as a fork of the area
        area_id_to_geo_id = {}

        for rule in self.rules:
            # Geographies are required in the spec
            if not rule["geographies"]:
                raise ValidationError(f"Rule {rule.rule_id} has no geographies.")

            for area_id in rule["geographies"]:
                geo_id = area_id_to_geo_id.get(area_id)
                if not geo_id:
                    geo_id = area_id_to_geo_id[area_id] = uuid.uuid4()

                if geo_id not in self.geographies:
                    try:
                        area = Area.objects.get(pk=area_id)
                    except Area.DoesNotExist:
                        raise ValidationError(
                            f"Geography {geo_id} referenced by "
                            f"rule {rule['rule_id']} is unknown."
                        )
                    # This is trigger a database query,
                    # just as a prefetch_related on a single area would anyway
                    if not len(area.polygons.all()):
                        raise ValidationError(f"Geography {geo_id} has no geometry.")

                    self.geographies[str(geo_id)] = {
                        "type": "Feature",
                        "geometry": {
                            "type": "GeometryCollection",
                            "geometries": [
                                json.loads(polygon.geom.geojson)  # Serialized
                                for polygon in area.polygons.all()
                            ],
                        },
                        "id": geo_id,  # In the GeoJSON spec
                        # Save as much as necessary from the original Area
                        "properties": {
                            # Not in the spec but a convention we already use
                            "name": area.label,
                            "label": area.label,
                            # Also keep a reference to the original area ID
                            "area": area_id,
                        },
                    }

            # Now swap the area IDs for geography IDs
            rule["geographies"] = list(
                map(area_id_to_geo_id.__getitem__, rule["geographies"])
            )

        if at is None:
            at = timezone.now()
        self.published_date = at

        self.save()


class Compliance(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    policy = models.ForeignKey(
        Policy, related_name="compliances", on_delete=models.CASCADE
    )
    vehicle = models.ForeignKey(
        Device, related_name="compliances", on_delete=models.CASCADE
    )
    rule = models.UUIDField()
    geography = models.UUIDField()
    start_date = models.DateTimeField()
    end_date = models.DateTimeField(blank=True, null=True)
    # Delay considered before computing the compliance
    lag = models.DurationField()
    # Remember to update this field within raw queries!
    saved_at = models.DateTimeField(db_index=True, auto_now=True)
    extra = pg_fields.JSONField(blank=True, null=True)

    class Meta:
        unique_together = (
            "policy",
            "rule",
            "geography",
            "vehicle",
            "start_date",
            # Different jobs with each one their lag follow a different timeline
            # and may compute different compliances as they may see more events
            "lag",
        )

    def __str__(self):
        return f"({short_uuid4(self.id)})"



import datetime
import random

from django.conf import settings
from django.contrib.gis.geos.point import Point
from django.utils.module_loading import import_string

from functools import lru_cache


def telemetry_is_enabled():
    """Default implementation for the ENABLE_TELEMETRY_FUNCTION setting."""
    return True


def is_telemetry_enabled():
    # Cache using the setting value as a key,
    # to get a different value from a setting override (tests)
    @lru_cache()
    def get_function_from_settings(setting):
        return import_string(setting)

    return get_function_from_settings(settings.ENABLE_TELEMETRY_FUNCTION)()


def to_mds_timestamp(dt: datetime.datetime) -> int:
    """
    Convert a datetime object into an MDS-compliant timestamp.

    MDS wants the timestamp to be in milliseconds.
    """
    return round(dt.timestamp() * 1000)


def from_mds_timestamp(value: int) -> datetime:
    """Convert an MDS timestamp (in milliseconds) to a datetime object."""
    return datetime.datetime.fromtimestamp(value / 1000, tz=datetime.timezone.utc)


def get_random_point(polygon):
    """Return a random point in the given polygon."""
    (x_min, y_min), (x_max, _), (_, y_max) = polygon.envelope[0][:3]
    while True:
        point = Point(random.uniform(x_min, x_max), random.uniform(y_min, y_max))
        if polygon.contains(point):
            return point



from rest_framework import exceptions
from rest_framework import response
from rest_framework import viewsets

from mds import models


class GeographyViewSet(viewsets.ReadOnlyModelViewSet):
    permission_classes = ()  # Public endpoint
    # Allow to access geographies from any published policy
    # Past or future, active or superseded by another policy
    queryset = models.Policy.objects.filter(published_date__isnull=False)

    def list(self, request, *args, **kwargs):
        # The spec excluded that use case
        raise exceptions.NotFound()

    def retrieve(self, request, *args, **kwargs):
        instance = self.get_object()
        # We already dump the geographies as a dict of FeatureCollections
        # Just flatten the dict to a list
        features = instance.geographies.values()

        return response.Response(
            # Geographies already stored in the GeoJSON format
            {"type": "FeatureCollection", "features": features}
        )



# Generated by Django 2.2.7 on 2020-01-06 17:15

from django.db import migrations, models


class Migration(migrations.Migration):
    atomic = False

    dependencies = [
        ("mds", "0002_pre_compliance_extra"),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            database_operations=[
                migrations.RunSQL(
                    """CREATE INDEX CONCURRENTLY IF NOT EXISTS "device_mds_events_partial" ON "mds_eventrecord" ("device_id") WHERE NOT ("event_type" = 'telemetry')"""
                )
            ],
            state_operations=[
                migrations.AddIndex(
                    model_name="eventrecord",
                    index=models.Index(
                        condition=models.Q(_negated=True, event_type="telemetry"),
                        fields=["device"],
                        name="device_mds_events_partial",
                    ),
                ),
            ],
        ),
    ]



from django.conf import settings

from cryptography import fernet


POLLER_TOKEN_CACHE = getattr(settings, "POLLER_TOKEN_CACHE", "default")
POLLER_TOKEN_ENCRYPTION_KEY = getattr(
    settings,
    "POLLER_TOKEN_ENCRYPTION_KEY",
    fernet.Fernet.generate_key(),  # The default is reset on each restart
)



from rest_framework import serializers
from rest_framework import viewsets

from django.db.models import Prefetch

from mds import models
from mds import utils


class ComplianceSerializer(serializers.ModelSerializer):
    rules_id = serializers.SerializerMethodField()
    compliances = serializers.SerializerMethodField()

    class Meta:
        model = models.Policy
        fields = ("id", "rules_id", "compliances")

    def get_rules_id(self, policy):
        return [rule["rule_id"] for rule in policy.rules]

    def get_compliances(self, policy):
        final_compliance = []
        for compliance in policy.compliances_pref:
            for current_compliance in final_compliance:
                if current_compliance["rule_id"] == compliance.rule:
                    for match in current_compliance["matches"]:
                        if match["geography"] == str(compliance.geography):
                            match["measured"] += 1
                            break
                    else:
                        current_compliance["matches"].append(
                            {"geography": str(compliance.geography), "measured": 1}
                        )
                    current_compliance["vehicles_in_violation"].append(
                        str(compliance.vehicle.id)
                    )
                    current_compliance["total_violations"] += 1
                    break
            else:
                final_compliance.append(
                    {
                        "rule_id": compliance.rule,
                        "matches": [
                            {"geography": str(compliance.geography), "measured": 1}
                        ],
                        "vehicles_in_violation": [str(compliance.vehicle.id)],
                        "total_violations": 1,
                    }
                )
        return final_compliance


class ComplianceViewSet(viewsets.ModelViewSet):
    queryset = models.Policy.objects.order_by("start_date")
    serializer_class = ComplianceSerializer

    def get_queryset(self):
        queryset = super().get_queryset()
        provider_id = self.request.GET.get("provider_id")
        end_date = self.request.GET.get("end_date")
        if end_date:
            end_date = utils.from_mds_timestamp(int(end_date))

        filters = {}
        if provider_id and end_date:
            filters["compliances__vehicle__provider__id"] = provider_id
            filters["compliances__start_date__lte"] = end_date
            filter_prefetch = Prefetch(
                "compliances",
                queryset=models.Compliance.objects.exclude(
                    end_date__lte=end_date
                ).filter(vehicle__provider__id=provider_id, start_date__lt=end_date),
                to_attr="compliances_pref",
            )
        elif end_date:
            filter_prefetch = Prefetch(
                "compliances",
                queryset=models.Compliance.objects.exclude(
                    end_date__lte=end_date
                ).filter(start_date__lt=end_date),
                to_attr="compliances_pref",
            )
            filters["compliances__start_date__lte"] = end_date
        elif provider_id:
            filter_prefetch = Prefetch(
                "compliances",
                queryset=models.Compliance.objects.filter(
                    vehicle__provider__id=provider_id
                ),
                to_attr="compliances_pref",
            )
            filters["compliances__vehicle__provider__id"] = provider_id
        else:
            filter_prefetch = Prefetch(
                "compliances",
                queryset=models.Compliance.objects.filter(),
                to_attr="compliances_pref",
            )

        if (provider_id and end_date) or end_date:
            queryset = (
                queryset.exclude(compliances__end_date__lt=end_date)
                .filter(**filters)
                .prefetch_related(filter_prefetch)
            )
        elif provider_id:
            queryset = queryset.filter(**filters).prefetch_related(filter_prefetch)
        else:
            queryset = queryset.prefetch_related(filter_prefetch)

        return queryset



# Generated by Django 2.2.7 on 2019-12-18 10:26

from django.conf import settings
import django.contrib.postgres.fields
from django.db import migrations, models
import django.db.models.deletion
import oauth2_provider.generators


class Migration(migrations.Migration):

    replaces = [
        ("authent", "0001_initial"),
        ("authent", "0002_auto_20190124_0928"),
        ("authent", "0003_pre_add_aggregator_for_application"),
        ("authent", "0004_post_fill_default_aggregator_for"),
        ("authent", "0005_pre_add_scope_restrictions_for_application_users"),
        ("authent", "0006_post_fill_allowed_scopes_with_default_values"),
        ("authent", "0007_post_make_application_user_scope_blankable"),
    ]

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("mds", "0001_squashed_0059_post_policy_config_encoder"),
    ]

    operations = [
        migrations.CreateModel(
            name="AccessToken",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                ("expires", models.DateTimeField()),
                ("scope", models.TextField(blank=True)),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                ("token", models.TextField()),
                ("jti", models.UUIDField(db_index=True)),
                ("revoked_after", models.DateTimeField(blank=True, null=True)),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="Application",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "client_id",
                    models.CharField(
                        db_index=True,
                        default=oauth2_provider.generators.generate_client_id,
                        max_length=100,
                        unique=True,
                    ),
                ),
                (
                    "redirect_uris",
                    models.TextField(
                        blank=True, help_text="Allowed URIs list, space separated"
                    ),
                ),
                (
                    "client_type",
                    models.CharField(
                        choices=[
                            ("confidential", "Confidential"),
                            ("public", "Public"),
                        ],
                        max_length=32,
                    ),
                ),
                (
                    "authorization_grant_type",
                    models.CharField(
                        choices=[
                            ("authorization-code", "Authorization code"),
                            ("implicit", "Implicit"),
                            ("password", "Resource owner password-based"),
                            ("client-credentials", "Client credentials"),
                        ],
                        max_length=32,
                    ),
                ),
                (
                    "client_secret",
                    models.CharField(
                        blank=True,
                        db_index=True,
                        default=oauth2_provider.generators.generate_client_secret,
                        max_length=255,
                    ),
                ),
                ("name", models.CharField(blank=True, max_length=255)),
                ("skip_authorization", models.BooleanField(default=False)),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                (
                    "owner",
                    models.UUIDField(
                        blank=True,
                        db_index=True,
                        help_text="Unique identifier for the owner of the application",
                        null=True,
                    ),
                ),
                (
                    "scopes",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=32),
                        help_text="Application allowed scopes (separated with commas)",
                        size=None,
                    ),
                ),
                (
                    "users_allowed_scopes",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=32),
                        blank=True,
                        default=list,
                        help_text="User's allowed scopes (separated with commas)",
                        size=None,
                    ),
                ),
                (
                    "aggregator_for",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Provider the application is allowed to write for",
                        to="mds.Provider",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="authent_application",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="RefreshToken",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                ("token", models.CharField(max_length=255)),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                ("revoked", models.DateTimeField(null=True)),
                (
                    "access_token",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="refresh_token",
                        to="authent.AccessToken",
                    ),
                ),
                (
                    "application",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="authent.Application",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="authent_refreshtoken",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={"abstract": False, "unique_together": {("token", "revoked")}},
        ),
        migrations.CreateModel(
            name="Grant",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                ("code", models.CharField(max_length=255, unique=True)),
                ("expires", models.DateTimeField()),
                ("redirect_uri", models.CharField(max_length=255)),
                ("scope", models.TextField(blank=True)),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                (
                    "application",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="authent.Application",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="authent_grant",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.AddField(
            model_name="accesstoken",
            name="application",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="authent.Application",
            ),
        ),
        migrations.AddField(
            model_name="accesstoken",
            name="source_refresh_token",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="refreshed_access_token",
                to="authent.RefreshToken",
            ),
        ),
        migrations.AddField(
            model_name="accesstoken",
            name="user",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="authent_accesstoken",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]
